"""Metadata pertaining to the xxkcd Python package"""

__author__ = 'Mital Ashok'
__credits__ = ['Mital Ashok']
__license__ = 'MIT'
__version__ = '1.4.0'
__maintainer__ = 'Mital Ashok'
__author_email__ = __email__ = 'mital.vaja@googlemail.com'
__status__ = 'Production'
__url__ = 'https://github.com/MitalAshok/xxkcd'
__description__ = 'An (unofficial) Python wrapper around xkcd APIs'



# coding: utf-8

import sys
import weakref
import json
import datetime
import random
import functools
import posixpath
import multiprocessing
import shutil
import contextlib

from objecttools import ThreadedCachedProperty

from xxkcd._util import (
    urlopen, reload, unescape, map, str_is_bytes, make_mapping_proxy,
    range, short, dead_weaklink, coerce_, index
)
from xxkcd import constants

__all__ = ('xkcd', 'load_xkcd_cache')

_404_mock = make_mapping_proxy({
    'month': 4, 'num': 404, 'link': '', 'year': 2008, 'news': u'',
    'safe_title': u'404 not found', 'transcript': u'', 'alt': u'', 'img': '',
    'title': u'404 - Not Found', 'day': 1
})


def _decode(s):
    u"""
    Decodes the incorrect encoding used in the xkcd JSON api.

    All of the fields of xkcd comics that use non-ascii characters incorrectly
    encode binary UTF-8 as escaped UTF-8.

    For example, 'é' -> 0xC3 0xA9 -> '\\u00c3\\u00a9' in the JSON, which
    actually encodes 'Ã©' (0xC3 0x83 0xC2 0xA9 in UTF-8)

    This is sometimes done multiple times, like in xkcd(124).

    :param str s: String to decode
    :return: Decoded string
    :rtype: Text
    """
    for _ in range(10):
        try:
            old, s = s, s.encode('latin-1').decode('utf-8')
            if old == s:
                break
        except (ValueError, UnicodeDecodeError):
            break
    return unescape(s)


# Python 2 and Python 3.6+ allow bytes JSON
_JSON_BYTES = sys.version_info < (3,) or sys.version_info >= (3, 6)

if not _JSON_BYTES:
    import codecs

    _UTF_8_READER = codecs.getreader('utf-8')

_MIMES = {
    'png': 'image/png', 'jpg': 'image/jpeg', 'gif': 'image/gif',
    'jpeg': 'image/jpeg'
}


___ = 'antigravity'
____ = sys.modules.get
_____ = None
______ = reload


def _loader(cls, n):
    return dict(cls(n)._raw_json)


_LAST_LATEST = 2128


class xkcd(object):
    """Interface with the xkcd JSON API"""
    __slots__ = ('_comic', '__weakref__', '__dict__')

    urlopen = staticmethod(urlopen)

    _cache = {}
    _keep_alive = {}

    def __new__(cls, comic=None, keep_alive=False):
        """
        Wrapper around the xkcd API for the comic

        :param Union[int, None, xkcd] comic: Number of the comic.
            -1 or `None` for the latest comic.
            -x means x comics before the latest comic.
        :param bool keep_alive: True to not delete the data requested until
            the `.delete()` method is called
        :return: The new xkcd object
        :rtype: xkcd
        :raises TypeError: Non-int given as comic
        """
        if isinstance(comic, xkcd):
            comic = comic.comic
        else:
            comic = coerce_(comic, cls.latest, _LAST_LATEST)
        self = cls._cache.get(comic, dead_weaklink)()
        if self is None:
            self = super(xkcd, cls).__new__(cls)
            self._comic = comic
            cls._cache[comic] = weakref.ref(self)
        if keep_alive:
            cls._keep_alive[comic] = self
        return self

    def __getnewargs__(self):
        return self.comic,

    def __reduce__(self):
        return type(self), (self.comic,)

    @staticmethod
    def with_opener(opener, name='xkcdWithCustomOpener', module=__name__, qualname=None, metaclass=type):
        """
        Takes an opener and returns an xkcd subclass that makes HTTP requests with that opener.

        Typical usage:

            # The __name__ ensures that the new class is picklable.
            xkcdA = xkcd.with_opener(opener, 'xkcdA', __name__)

        The opener is a callable object (e.g. a type). When called with a `str` for the
        url, it should return an object with the following methods:
            `read(n: int)`: Returns that many bytes from the response from the url.
            `read()`: Read from the request and return as many bytes as possible.
            `close()`: Closes the connection and no more methods will be called before open.

        There may be more than one open connection at a time.

        For example:

            import requests

            class Opener(object):
                __slots__ = ('response',)

                _session = None

                @classmethod
                def get(cls, url):
                    if cls._session is None:
                        cls._session = requests.Session()
                    return cls._session.get(url)

                def __init__(self, url):
                    self.response = self.get(url)

                def read(self, n=None):
                    if n is None:
                        return self.response.content
                    return self.response.raw.read(n)

                def close(self):
                    self.response.close()

        :param opener: The opener, as described above.
        :param str name: The name of the subclass.
        :param str module: The module of the new type.
        :param Optional[str] qualname: The qualified name of the class.
          Defaults to `module + '.' + name`.
        :param type metaclass: The metaclass of the new object.
          The new class is constructed as `metaclass(name, (xkcd,), <class __dict__>)`.
        :return: A new subclass of `xkcd` with a custom `.urlopen` staticmethod.
        """
        @staticmethod
        def urlopen(url):
            return contextlib.closing(opener(url))

        d = {
          '__slots__': xkcd.__slots__,
          'urlopen': urlopen,
          '_cache': {},
          '_keep_alive': {},
          '__module__': module
        }

        if hasattr(object, '__qualname__'):
            if qualname is None:
                qualname = module + '.' + name
            d['__qualname__'] = qualname
            urlopen.__qualname__ = qualname + '.urlopen'
        return metaclass(name, (xkcd,), d)

    @property
    def comic(self):
        return self._comic

    @ThreadedCachedProperty
    def _raw_json(self):
        """Raw JSON with a possibly incorrect transcript and alt text"""
        if self.comic == 404:
            return make_mapping_proxy(_404_mock)
        if self.comic is None:
            url = constants.xkcd.json.latest
        else:
            url = constants.xkcd.json.for_comic(number=self.comic)
        with self.urlopen(url) as http:
            if not _JSON_BYTES:
                http = _UTF_8_READER(http)
            return make_mapping_proxy(json.load(http))

    _raw_json.can_delete = True
    _raw_json.can_set = True

    @ThreadedCachedProperty
    def json(self):
        """
        JSON with the correct transcript and encoded properly.

        Note, that in Python 2, all of the `str` keys are `unicode`.

        Format::
        {
            'month': (the number of the month it was published. Possibly None.),
            'num': (int of the number of the comic),
            'link': (str of where the comic links to when clicked on, most commonly a larger version),
            'year': (the year it was published. Possibly None.),
            'news': (str of news that would be placed on the news ticker on the comic. Usually empty.),
            'safe_title': (str of url safe title),
            'transcript': (str of the transript. Possibly empty, usually for later comics),
            'alt': (str of the text that appears in the title attribute of the comic. Possibly empty.),
            'img': (str of the direct link to the image. Possibly empty, usually for interactive comics),
            'title': (str of the original title of the comic.),
            'day': (the number of the day it was published. Possibly None.)
        }

        Type stub::
        {
            'month': Optional[int], 'num': int, 'link': Text,
            'year': Optional[int], 'news': Text, 'safe_title': Text,
            'transcript': Text, 'alt': Text, 'img': Text, 'title': Text,
            'day' Optional[int]
        }
        """
        n = self.comic
        if n is None:
            pass
        # These comics messed up the transcripts.
        # They were interactive comics that didn't have a transcript that
        # pushed the transcripts for all other comics 1 or 2 ahead.
        elif n >= 1663:
            n += 3
        elif n >= 1608:
            n += 2
        if str_is_bytes:
            other_json = {}
            raw_json = self._raw_json
            for key in raw_json:
                other_json[key.encode('ascii')] = raw_json[key]
        else:
            other_json = dict(self._raw_json)
        if n is not None and n != self.comic and (n < _LAST_LATEST - 3 or n < self.latest() - 3):
            other_json['transcript'] = _decode(xkcd(n)._raw_json['transcript'])
        else:
            other_json['transcript'] = _decode(other_json['transcript'])
        for text_key in ('alt', 'title', 'safe_title'):
            other_json[text_key] = _decode(other_json[text_key])
        for int_key in ('day', 'month', 'year'):
            if other_json[int_key]:
                other_json[int_key] = short(other_json[int_key])
            else:
                other_json[int_key] = None
        if other_json['img'] == constants.xkcd.images.blank:
            other_json['img'] = ''
        if str_is_bytes:
            other_json['img'] = other_json['img'].encode('ascii')
            other_json['link'] = other_json['link'].encode('ascii')
        return make_mapping_proxy(other_json)

    json.can_delete = True

    @property
    def month(self):
        return self.json['month']

    @property
    def link(self):
        return self.json['link']

    @property
    def year(self):
        return self.json['year']

    @property
    def news(self):
        return self.json['news']

    @property
    def safe_title(self):
        return self.json['safe_title']

    @property
    def transcript(self):
        return self.json['transcript']

    @property
    def alt(self):
        return self.json['alt']

    @property
    def img(self):
        return self.json['img']

    def read_image(self):
        """
        Read the raw image from the link and return the data.

        :raises ValueError: Comic doesn't have an image (e.g., it is xkcd(404), or it is interactive.)
        :return: Raw image data
        :rtype: bytes
        """
        if not self.img:
            raise ValueError('Comic {} does not have an image!'.format(self))
        with self.urlopen(self.img) as http:
            # http.read may not read all at once.
            return b''.join(iter(http.read, b''))

    def stream_image(self, file=None, chunk_size=16384):
        """
        Stream the raw image file from the link and write to file.

        :param file: File-like to write to, or None to return an iterator over the chunks
        :type file: Optional[BinaryIO]
        :param chunk_size: Size of chunks (or None for the whole file in a single chunk. Consider `self.read_image()`.)
        :type chunk_size: Optional[int]
        :return: iterator or None
        :rtype: Optional[Iterator[bytes]]
        """
        if not self.img:
            raise ValueError('Comic {} does not have an image!'.format(self))
        opened = self.urlopen(self.img)
        if file is None:
            if chunk_size is None:
                with opened as http:
                    return iter((b''.join(iter(http.read, b'')),))

            def _generator():
                # Use an inner function so the function
                # can end normally if file is not None
                with opened as http:
                    read = functools.partial(http.read, chunk_size)
                    for chunk in iter(read, b''):
                        yield chunk
            return _generator()
        else:
            with opened as http:
                if chunk_size is None:
                    file.write(b''.join(iter(http.read, b'')))
                else:
                    shutil.copyfileobj(http, file, chunk_size)

    @property
    def image_filename(self):
        """Suggested name for the image file. None if no image."""
        if not self.img:
            return None
        return posixpath.basename(self.img)

    @property
    def image_ext(self):
        """File extension of image file. Usually '.png'. '' if no extension, None if no image."""
        image_name = self.image_filename
        if image_name is None:
            return None
        return posixpath.splitext(image_name)[1]

    @property
    def image_name(self):
        """`self.image_filename` without the extension."""
        image_name = self.image_filename
        if image_name is None:
            return None
        return posixpath.splitext(image_name)[0]

    @property
    def image_mime(self):
        """
        Mime type of the image. Usually 'image/png'.
        'application/octet-stream' if not found, None if no image.
        """
        ext = self.image_ext
        if ext is None:
            return None
        return _MIMES.get(ext.lstrip('.').lower(), 'application/octet-stream')

    @property
    def title(self):
        return self.json['title']

    @property
    def day(self):
        return self.json['day']

    @property
    def date(self):
        """
        :return: The day, month and year the comic was published.
        :rtype: datetime.date
        """
        return datetime.date(self.year, self.month, self.day)

    @classmethod
    def latest(cls):
        """
        :return: The number of the latest comic
        :rtype: int
        """
        return cls(keep_alive=True)._raw_json['num']

    @classmethod
    def random(cls):
        """
        :return: A random comic
        :rtype: xkcd
        """
        return cls(random.randint(1, cls.latest()))

    def delete(self):
        """
        Deletes the data associated with this xkcd object so that it is reloaded
        the next time it is requested

        :return: None
        """
        del self._raw_json
        del self.json
        self._keep_alive.pop(self.comic, None)
        self._cache.pop(self.comic, None)

    @property
    def explain_xkcd(self):
        """
        :return: The explain xkcd wiki link for this comic
        :rtype: str
        """
        if self.comic is None:
            return constants.explain_xkcd.latest
        return constants.explain_xkcd.for_comic(number=self.comic)

    @property
    def url(self):
        """
        :return: The url for this comic
        :rtype: str
        """
        if self.comic is None:
            return constants.xkcd.latest
        return constants.xkcd.for_comic(number=self.comic)

    @property
    def mobile_url(self):
        """
        :return: The mobile url for this comic
        :rtype: str
        """
        if self.comic is None:
            return constants.xkcd.mobile.latest
        return constants.xkcd.mobile.for_comic(number=self.comic)

    @property
    def num(self):
        return self.json['num']

    def __repr__(self):
        n = self.comic
        n = '' if n is None else n
        return '{type.__name__}({number})'.format(type=type(self), number=n)

    @classmethod
    def load_all(cls, processes=None):
        """
        Load all the comics into the cache. Note: Takes a lot of time.

        :param Optional[int] multiprocessed: Number of processes to use
            in parallel. If `None`, use only the current process.
        :return: None
        """
        if processes is not None:
            pool = multiprocessing.Pool(processes)
            data = pool.map(cls._get_loader(), (i for i in cls.range() if not cls._raw_json.is_cached(cls(i, keep_alive=True))))
            pool.close()
            pool.join()
            for n, raw_json in enumerate(data, 1):
                cls(n, keep_alive=True)._raw_json = make_mapping_proxy(raw_json)
        else:
            for n in cls.range():
                cls(n, keep_alive=True)._raw_json
        # Load last comic
        cls(keep_alive=True)._raw_json

    # Python 2 doesn't allow pickling classmethod for multiprocessing.
    if sys.version_info >= (3,):
        _loader = classmethod(_loader)

        @classmethod
        def _get_loader(cls):
            return cls._loader
    else:
        @classmethod
        def _get_loader(cls):
            return functools.partial(_loader, cls)

    @classmethod
    def load_one(cls, n):
        cls(n, keep_alive=True)._raw_json

    @classmethod
    def delete_all(cls):
        cls._keep_alive.clear()
        for comic in cls._cache.values():
            comic = comic()
            if comic is not None:
                comic.delete()

    @classmethod
    def delete_one(cls, comic):
        comic = index(comic)
        cls._keep_alive.pop(comic, None)
        comic = cls._cache.get(comic, dead_weaklink)()
        if comic is not None:
            comic.delete()

    @staticmethod
    def antigravity():
        """???"""
        antigravity = ____(___)
        if antigravity is not _____:
            return ______(antigravity)
        import antigravity
        return antigravity

    def __iter__(self):
        if self.comic is None:
            return iter((self,))
        cls = type(self)
        return map(cls, range(self.comic, cls.latest() + 1))

    def next(self, keep_alive=False):
        """
        Return the next comic. If is the latest comic, raise StopIteration

        :return: Next comic
        :rtype: xkcd
        """
        return self.__next__(keep_alive)

    def back(self, keep_alive=False):
        """
        Return the previous comic. If is xkcd(1), raise StopIteration.

        :return: Previous comic
        :rtype: xkcd
        """
        if self.comic is None:
            return type(self)(self.latest() - 1, keep_alive)
        if self.comic == 1:
            raise StopIteration
        return type(self)(self.comic - 1, keep_alive)

    def __next__(self, keep_alive=False):
        if self.comic is None or (self.comic >= _LAST_LATEST and self.comic == self.latest()):
            raise StopIteration
        return type(self)(self.comic + 1, keep_alive)

    def __reversed__(self):
        if self.comic is None:
            comic = self.latest()
        else:
            comic = self.comic
        cls = type(self)
        return iter(map(cls, range(comic - 1, 0, -1)))

    @classmethod
    def range(cls, from_=1, to=None, step=1):
        """
        Returns an iterator over specified comics.

        Use negative numbers to specify from the end.

        :param int from_: The comic to start from. Defaults to the first.
        :param Optional[int] to: The comic to end by (exclusive). Defaults to last + 1.
        :param int step: The step amount. Defaults to one.
        :return: A range over the requested comics.
        """
        if to is None:
            if step < 0:
                to = 0
            else:
                to = cls.latest() + 1

        return range(from_, to, step)

    def __eq__(self, other):
        if isinstance(other, xkcd) and isinstance(self, xkcd):
            if self is other or other.comic == self.comic:
                return True
            if self.comic is None:
                return self.num == other.comic
            if other.comic is None:
                return other.num == self.comic
            return False
        return NotImplemented

    def __ne__(self, other):
        eq = self.__eq__(other)
        if eq is NotImplemented:
            return NotImplemented
        return not eq

    def __lt__(self, other):
        if isinstance(other, xkcd) and isinstance(self, xkcd):
            if self is other or self.comic == other.comic:
                return False
            if self.comic is None:
                return False
            if other.comic is None:
                return True
            return self.comic < other.comic
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, xkcd) and isinstance(self, xkcd):
            if self is other or self.comic == other.comic:
                return True
            if self.comic is None:
                if other.comic is None:
                    return True
                return self.num == other.comic
            if other.comic is None:
                return True
            return self.comic <= other.comic
        return NotImplemented

    def __ge__(self, other):
        lt = self.__lt__(other)
        if lt is NotImplemented:
            return NotImplemented
        return not lt

    def __gt__(self, other):
        le = self.__le__(other)
        if le is NotImplemented:
            return NotImplemented
        return not le


def load_xkcd_cache():
    """
    Loads a pre-fetched cache of the xkcd API. Currently for 1 through 2128.
    Might be a bit stale. Will still make HTTP requests for newer comics.
    """
    from xxkcd._cache import cache

    for k, v in getattr(dict, 'iteritems', dict.items)(cache):
        xkcd(k, keep_alive=True)._raw_json = make_mapping_proxy(v)



import datetime
import collections
import weakref
import random

from objecttools import ThreadedCachedProperty

from xxkcd import constants
from xxkcd._util import urlopen, make_mapping_proxy, range, str_is_bytes, coerce_, dead_weaklink
from xxkcd._html_parsing import ParseToTree

__all__ = ('WhatIf',)

ArchiveEntry = collections.namedtuple('ArchiveEntry', ('image', 'title', 'date'))
ArticlePart = collections.namedtuple('ArticlePart', ('type', 'value'))


class Archive(object):
    _archive = None
    _months = [
        'January', 'February', 'March', 'April', 'May', 'June', 'July',
        'August', 'September', 'October', 'November', 'December'
    ]

    @ThreadedCachedProperty
    def _length(self):
        return len(self.__get__())

    _length.can_delete = True

    def __get__(self, instance=0, owner=None):
        if instance is None:
            return self
        archive = Archive._archive
        if archive is not None:
            return make_mapping_proxy(archive)
        parser = ParseToTree()
        with urlopen(constants.what_if.archive) as http:
            data = http.read()
        tree = parser(data)
        archive = {}
        for i, archive_entry in enumerate(tree.find_all(Archive._is_archive_entry), 1):
            c = archive_entry.element_children
            image = constants.what_if.base + c[0].first_element_child.attr_dict['src']
            if str_is_bytes:
                image = image.encode('ascii')
            archive[i] = ArchiveEntry(
                image=image,
                title=c[1].first_element_child.children[0].children,
                date=self._parse_date(c[2].children[0].children)
            )
        article = Archive._archive = make_mapping_proxy(archive)
        return make_mapping_proxy(article)

    def __delete__(self, instance):
        Archive._archive = None
        del Archive._length

    def __len__(self):
        return self._length

    @staticmethod
    def _is_archive_entry(node):
        return node.cls == 'archive-entry'

    @classmethod
    def _parse_date(cls, date):
        month, day, year = date.split()
        month = cls._months.index(month)
        day = int(day[:-1])
        year = int(year)
        return datetime.date(year, month + 1, day)


_LAST_LATEST = 157


class WhatIf(object):
    __slots__ = ('_article', '__weakref__', '__dict__')

    _cache = {}
    _keep_alive = {}

    archive = Archive()

    def __new__(cls, article=None, keep_alive=False):
        if isinstance(article, cls):
            if keep_alive:
                cls._keep_alive[article.article] = article
            return article
        article = coerce_(article, cls.latest, _LAST_LATEST)
        self = cls._cache.get(article, dead_weaklink)()
        if self is None:
            self = super(WhatIf, cls).__new__(cls)
            self._article = article
            cls._cache[article] = weakref.ref(self)
        if keep_alive:
            cls._keep_alive[article] = self
        return self

    @classmethod
    def latest(cls):
        return len(cls.archive)

    @classmethod
    def random(cls):
        return WhatIf(random.randint(1, cls.latest()))

    @property
    def article(self):
        return self._article

    @property
    def number(self):
        if self._article is None:
            return self.latest()
        return self._article

    @property
    def image(self):
        return self.archive[self.number].image

    img = image  # To remain consistent with the xkcd API

    @property
    def title(self):
        return self.archive[self.number].title

    @property
    def date(self):
        return self.archive[self.number].date

    @property
    def month(self):
        return self.date.month

    @property
    def day(self):
        return self.date.day

    @property
    def year(self):
        return self.date.year

    @staticmethod
    def news():
        with urlopen(constants.xkcd.c.what_if.news) as http:
            return http.read()

    @property
    def url(self):
        if self.article is None:
            return constants.what_if.latest
        return constants.what_if.for_article(number=self.article)

    @ThreadedCachedProperty
    def full_page(self):
        with urlopen(self.url) as http:
            return http.read().decode('utf-8')

    full_page.can_delete = True

    @ThreadedCachedProperty
    def _article_tree(self):
        tree = ParseToTree()(self.full_page)
        return tree.find(lambda node: node.tag.lower() == 'article' and node.cls == 'entry')

    _article_tree.can_delete = True

    @ThreadedCachedProperty
    def _article_node(self):
        return self._article_tree.find(lambda a: a.cls == 'entry')

    _article_node.can_delete = True

    @ThreadedCachedProperty
    def question(self):
        return str(self._article_tree.get_element_by_id('question').children[0])

    question.can_delete = True

    @ThreadedCachedProperty
    def attribute(self):
        return str(
            self._article_tree.get_element_by_id('attribute').children[0]
        )

    attribute.can_delete = True

    @ThreadedCachedProperty
    def body(self):
        return str(self._article_node)

    body.can_delete = True

    def delete(self, remove_cache=True):
        del self.full_page
        del self._article_tree
        del self._article_node
        del self.question
        del self.attribute
        del self.body
        if remove_cache:
            self._keep_alive.pop(self.article, None)
            self._cache.pop(self.article, None)

    def refresh(self):
        self.delete(False)
        self.full_page
        self._article_tree
        self._article_node
        self.question
        self.attribute
        self.body

    def next(self):
        return self.__next__()

    def back(self):
        if self.article is None:
            return type(self)(self.latest() - 1)
        if self.article == 1:
            raise StopIteration
        return type(self)(self.article - 1)

    def __next__(self):
        if self.article is None or (self.article >= _LAST_LATEST and self.article == self.latest()):
            raise StopIteration
        return type(self)(self.article + 1)

    def __reversed__(self):
        if self.article is None:
            comic = self.latest()
        else:
            comic = self.article
        return iter(map(type(self), range(comic - 1, 0, -1)))

    @classmethod
    def range(cls, from_=1, to=None, step=1):
        if to is None:
            to = cls.latest() + 1
        return range(from_, to, step)

    def __eq__(self, other):
        if isinstance(other, WhatIf) and isinstance(self, WhatIf):
            if self is other or other.article == self.article:
                return True
            if self.article is None:
                return self.number == other.article
            if other.article is None:
                return other.number == self.article
            return False
        return NotImplemented

    def __ne__(self, other):
        eq = self.__eq__(other)
        if eq is NotImplemented:
            return NotImplemented
        return not eq

    def __lt__(self, other):
        if isinstance(other, WhatIf) and isinstance(self, WhatIf):
            if self is other or self.article == other.article:
                return False
            if self.article is None:
                return False
            if other.article is None:
                return True
            return self.article < other.article
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, WhatIf) and isinstance(self, WhatIf):
            if self is other or self.article == other.article:
                return True
            if self.article is None:
                if other.article is None:
                    return True
                return self.number == other.article
            if other.article is None:
                return True
            return self.article <= other.article
        return NotImplemented

    def __ge__(self, other):
        lt = self.__lt__(other)
        if lt is NotImplemented:
            return NotImplemented
        return not lt

    def __gt__(self, other):
        le = self.__le__(other)
        if le is NotImplemented:
            return NotImplemented
        return not le

    def __repr__(self):
        n = self.article
        n = '' if n is None else n
        return '{type.__name__}({article})'.format(type=type(self), article=n)



__all__ = ('xkcd', 'explain_xkcd', 'what_if')

SECURE_PROTOCOL = 'https:'
INSECURE_PROTOCOL = 'http:'


class Constantsxkcd(object):
    base = SECURE_PROTOCOL + '//xkcd.com'
    with_subdomain = (SECURE_PROTOCOL + '//{subdomain}.xkcd.com').format
    latest = base
    for_comic = (base + '/{number}/').format


class ConstantsxkcdJSON(object):
    base = Constantsxkcd.base
    suffix = '/info.0.json'
    latest = base + suffix
    for_comic = (base + '/{number}' + suffix).format
    # Or, alternatively, `http://c.xkcd.com/api-0/jsonp/comic/{number}`


class ConstantsxkcdC(object):
    base = Constantsxkcd.with_subdomain(subdomain='c')


class ConstantsxkcdCWhatIf(object):
    base = ConstantsxkcdC.base + '/whatif'
    news = base + '/news'


class ConstantsxkcdMobile(object):
    base = Constantsxkcd.with_subdomain(subdomain='m')
    latest = base
    for_comic = (base + '/{number}/').format


class ConstantsxkcdImages(object):
    base = Constantsxkcd.with_subdomain(subdomain='imgs') + '/comics'
    for_image = (base + '/{image}').format
    blank = base + '/'


class ConstantsExplainxkcd(object):
    base = INSECURE_PROTOCOL + '//www.explainxkcd.com'
    latest = base
    for_comic = (base + '/{number}').format


class ConstantsWhatIf(object):
    base = Constantsxkcd.with_subdomain(subdomain='what-if')
    archive = base + '/archive/'
    latest = base
    for_article = (base + '/{number}/').format


xkcd = Constantsxkcd()
xkcd.json = ConstantsxkcdJSON()
xkcd.c = ConstantsxkcdC()
xkcd.c.what_if = ConstantsxkcdCWhatIf()
xkcd.mobile = ConstantsxkcdMobile()
xkcd.images = ConstantsxkcdImages()

explain_xkcd = ConstantsExplainxkcd()

what_if = ConstantsWhatIf()



"""An (unofficial) Python wrapper around xkcd APIs"""

from xxkcd.metadata import *
from xxkcd.xkcd import xkcd, load_xkcd_cache
from xxkcd.what_if import WhatIf

__all__ = ('xkcd', 'load_xkcd_cache', 'WhatIf')



#!/usr/bin/env python

import sys
import os
import codecs

from setuptools import setup, find_packages


__dir__ = os.path.abspath(os.path.dirname(__file__))

# To prevent a redundant __version__, import it from the packages
sys.path.insert(0, os.path.join(__dir__, 'xxkcd'))

try:
    from metadata import (
        __version__, __author__, __email__, __license__, __url__,
        __description__
    )
finally:
    sys.path.pop(0)

with codecs.open(os.path.join(__dir__, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup_args = dict(
    name='xxkcd',

    version=__version__,

    description=__description__,
    long_description=long_description,

    url=__url__,

    author=__author__,
    author_email=__email__,

    license=__license__,

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
    ],
    platforms=['any'],

    keywords=['xkcd', 'api', 'wrapper', 'what-if'],

    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'examples']),

    install_requires=[
        'objecttools>=1.0.1'
    ],
    extras_require={},
    entry_points={},

    test_suite='tests'
)


setup(**setup_args)



# Compatibility

try:
    import builtins
except ImportError:
    import __builtin__ as builtins

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen as _urlopen
    import contextlib

    def urlopen(*args, **kwargs):
        return contextlib.closing(_urlopen(*args, **kwargs))

try:
    from html.parser import HTMLParser
except ImportError:
    from HTMLParser import HTMLParser

try:
    from types import MappingProxyType
except ImportError:
    try:
        from types import DictProxyType as MappingProxyType
    except ImportError:
        MappingProxyType = None

try:
    MappingProxyType({'a': None})['a']
except:
    MappingProxyType = None

if MappingProxyType is None:
    class MappingProxyType(object):
        __slots__ = ('_m')

        def __init__(self, mapping):
            self._m = mapping

        def __cmp__(self, other):
            return cmp(self._m, other)

        def __contains__(self, key):
            return key in self._m

        def __eq__(self, other):
            return self._m == other

        def __ge__(self, other):
            return self._m > other

        def __getitem__(self, key):
            return self._m[key]

        def __gt__(self, other):
            return self._m > other

        def __iter__(self):
            return iter(self._m)

        def __le__(self, other):
            return self._m <= other

        def __len__(self):
            return len(self._m)

        def __lt__(self, other):
            return self._m < other

        def __ne__(self, other):
            return self.m != other

        def __repr__(self):
            return '{}({!r})'.format(type(self).__name__, self._m)

        def copy(self):
            return self._m.copy()

        def get(self, key, default=None):
            return self._m.get(key, default)

        def has_key(self, key):
            return self._m.has_key(key)

        def items(self):
            return self._m.items()

        def iteritems(self):
            return self._m.iteritems()

        def iterkeys(self):
            return self._m.iterkeys()

        def itervalues(self):
            return self._m.itervalues()

        def keys(self):
            return self._m.keys()

        def values(self):
            return self._m.values()


def make_mapping_proxy(mapping):
    if isinstance(mapping, MappingProxyType):
        return mapping
    return MappingProxyType(mapping)


try:
    from html import unescape  # Python 3.4+
except ImportError:
    unescape = HTMLParser().unescape

try:
    from html import escape
except ImportError:
    def escape(s, quote=True):
        s = s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')
        if quote:
            return s.replace('"', '&quot;').replace("'", '&#x27;')
        return s


text_type = type(u'')
str_type = type('')
binary_type = type(b'')

str_is_unicode = str_type is text_type
str_is_bytes = str_type is binary_type

try:
    from importlib import reload
except ImportError:
    reload = builtins.reload

try:
    from itertools import imap as map, ifilter as filter, izip as zip
except ImportError:
    map = builtins.map
    zip = builtins.zip
    filter = builtins.filter

try:
    from functools import reduce
except ImportError:
    reduce = builtins.reduce

range = getattr(builtins, 'xrange', range)
int = getattr(builtins, 'long', int)
short = builtins.int

try:
    from operator import index
except ImportError:
    def index(o):
        return range(o)[-1] + 1  # Use xrange to coerce to integer

try:
    infinity = float('inf')
except ValueError:
    infinity = 1e999 ** 2
neg_infinity = -infinity


def dead_weaklink():
    return None


def coerce_(x, max_fn, last_latest):
    if x is None:
        return x
    try:
        x = int(index(x))
    except (ValueError, TypeError):
        try:
            x = float(x)
            if x >= infinity:
                return None
        except (ValueError, TypeError):
            pass
    else:
        try:
            x = short(x)
        except Exception:
            pass
        if x == 0:
            return None
        if x <= 0:
            x = max_fn() + 1 + x
            if x <= 0:
                return 1
            return x
        if x <= last_latest:
            return x
        return min(x, max_fn())
    raise TypeError('comic must be an integer')



import collections

import objecttools

from xxkcd._util import HTMLParser, text_type, map, escape


@objecttools.Singleton.as_decorator
class TextNode(str):
    def __new__(cls):
        return str.__new__(TextNode, '<text>')

    def __repr__(self):
        return 'text_node'

text_node = TextNode()


class HTMLNode(
    collections.namedtuple('HTMLNode', ('tag', 'attrs', 'children', 'parent'))
):
    def find(self, criteria):
        return next(self.find_all(criteria), None)

    def find_all(self, criteria):
        queue = collections.deque()
        queue.append(self)
        while queue:
            current = queue.popleft()
            if criteria(current):
                yield current
            if current.tag is not text_node:
                queue.extend(current.children)

    def root(self):
        node = self
        while node.parent is not None:
            node = node.parent
        return node

    @property
    def cls(self):
        return self.attr_dict.get('class', '')

    @property
    def first_element_child(self):
        for child in self.children:
            if child.tag is not text_node:
                return child
        return None

    @property
    def element_children(self):
        return [child for child in self.children if child.tag is not text_node]

    @property
    def attr_dict(self):
        return dict(reversed(self.attrs))

    @property
    def id(self):
        return self.attr_dict.get('id', None)

    def get_element_by_id(self, id):
        return self.find(lambda node: node.id == id)

    def __repr__(self):
        if self.tag is text_node:
            return '<TextNode {!r}>'.format(self.children)
        if self.parent is None:
            parent = None
        elif self.parent.tag is text_node:
            parent = repr(self.parent)
        else:
            parent = '<{}/>'.format(self.parent.tag)
        children = ', '.join(
            '<{}/>'.format(n.tag) if n.tag is not text_node else repr(n)
            for n in self.children
        )
        return (
            '<HTMLNode tag={self.tag} attrs={self.attrs} '
            'children=[{children}] parent={parent}>'
        ).format(self=self, children=children, parent=parent)

    def __str__(self):
        if self.tag is text_node:
            return self.children
        if self.attrs:
            attrs = ' ' + ' '.join('{}="{}"'.format(attr, escape(value)) for attr, value in self.attrs)
        else:
            attrs = ''
        return '<{self.tag}{attrs}>{contents}</{self.tag}>'.format(
            self=self, attrs=attrs,
            contents=u''.join(map(text_type, self.children))
        )




class ParseToTree(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self._entry = self.tree = HTMLNode('_root', [], [], None)

    def handle_starttag(self, tag, attrs):
        new_entry = HTMLNode(tag.lower(), attrs, [], self._entry)
        self._entry.children.append(new_entry)
        self._entry = new_entry

    def handle_endtag(self, tag):
        while self._entry.tag != tag.lower():
            self._entry = self._entry.parent
        self._entry = self._entry.parent

    def handle_data(self, data):
        self._entry.children.append(HTMLNode(text_node, [], data, self._entry))

    def reset(self):
        self._entry = self.tree = HTMLNode('_root', [], [], None)
        HTMLParser.reset(self)

    def __call__(self, text):
        self.reset()
        if not isinstance(text, text_type):
            text = text.decode('utf-8')
        self.feed(text)
        self.close()
        parsed = self.tree
        self.reset()
        return parsed
